/*
 *  btgconst.h      Header file for btree.c and isam.c
 *
 *
 *              There are detailed comments above all the paramaters that
 *              can or should be tailored. Please take a few minutes and
 *              read through this file; all troubles getting the package
 *              working initially are due to one or more of these #defines
 *              being incorrectly set.
 *
 *                  fcntl.h         Some compilers require this header be
 *                                  #included for file access routines. 
 *                                  If yours is not one of them, comment out
 *                                  the #include.
 *
 *                  CREATMODE       The argument passed to creat()
 *
 *                  RWMODE          The argument passed to open()
 *          
 *                  voids           If your compiler does NOT support voids,
 *                                  uncomment the 'void' #define below.
 *
 *                  MAXKEYSIZE      The maximum key length. Don't
 *                                  forget the trailing NULL and the
 *                                  extra 3 byte overhead per key for
 *                                  duplicates.
 *
 *                  NUMINDEXS       The maximum number of index files that
 *                                  can be open at one time.
 *
 */

/*
 *  fcntl.h
 *
 *  Some compilers have a header file that must be included for file
 *  input/output functions. If your compiler comes with a fcntl.h file,
 *  then uncomment this #include
 */
#include    <stdio.h>
#include    <stdlib.h>
#include    <fcntl.h>
#include    <memory.h>
/*
 *  Some common CREATMODE and RWMODE values
 *  ---- ------ --------- --- ------ ------
 *
 *  Wizard C            CREATMODE  == 0200
 *                      RWMODE     == O_RDWR
 *
 *  MicroSoft v3.x      CREATMODE  == 0200
 *                      RWMODE     == (O_RDWR | O_BINARY)
 *                      (sign extends chars)
 *
 *  Hi-Tech C           CREATMODE  == 0200   (not really critical)
 *                      RWMODE     == 2
 *                      (doesn't require fcntl.h #include)
 *
 *  Lattice 3.0+        CREATMODE == (0x8002 | S_IWRITE | S_IREAD)
 *                      RWMODE    == 0x8002
 *
 *  Mark Williams C     CREATMODE == 0200 (doesn't matter)
 *                      RWMODE    == 2
 *
 *  DataLight C         CREATMODE  == 0
 *                      RWMODE     == 2
 *
 *  CWare DeSmet C      CREATMODE  == 0644
 *                      RWMODE     == 2
 *                      (sign extends chars)
 *
 *  Lattice < 3.0       CREATMODE  == 0x8644
 *                      RWMODE     == 0x8002
 *
 *  C.I. C86            CREATMODE  == BUPDATE
 *                      RWMODE     == BUPDATE
 *
 *  unix/xenix          CREATMODE  == 0644
 *                      RWMODE     == 2
 */
#define    CREATMODE    0200
#define    RWMODE       (O_RDWR | O_BINARY)
/*
 *  Voids
 *  -----
 *  if your compiler supports void types, comment out the void #define.
 *  Most compilers now do.
 */
#define VOID
/*
 *  other values that can be tailored
 */
#define    MAXKEYSIZE   20             /* max # of chars in key (less null)  */
#define    NUMINDEXS    10             /* max # of open files (key and data) */
#define    MINKEYS      31             /* b-tree order                       */
/*
 *  end of stuff that should be changed
 */

/*
 *      General purpose Symbolic constants:
 */
#define OK 0            /* General purpose "no error" return value          */
#define NAMESIZE  65    /* maximum path/file name length                    */
#define SECSIZ 128      /* Sector size for header read/write calls          */
#define LOCAL static    /* f'ns local to this module                        */
#define ENTRY           /* entry points into this module                    */
#define MAXTREEHEIGHT 10   /* huge height 10 tree (10^10 keys)              */

/*
 *  structure for lioGet/PutCur(). These are currently undocumented
 *  entry points that set and restore BTree currency.
 */
struct currency {
        short sPtr;
        long stk[MAXTREEHEIGHT];
        char last[MAXKEYSIZE];
    };

/*
 *  #defines that may already be defined in stdio.h
 */

/*
 *  NULL is used as a non-success return code
 */
#ifndef NULL
#include    <stdio.h>
#endif

/*
 *  many BTree routines need a (long) NULL
 */
#ifndef LNULL
#define LNULL (long) NULL
#endif

/*
 *  file position equates for btFPos() (only used in isam)
 */
#ifndef EOF
#define EOF (-1)        /* end of file       */
#endif
#define BOF (-2)        /* beginning of file */
#define WHOKNOWS (-3)    /* unknown position  */

/*
 *  general "everything's OK" value
 */
#ifndef GOOD
#define GOOD 1
#endif

/*
 *  life threatening error value
 */
#ifndef ERROR
#define ERROR -1
#endif

/*
 *  true/false
 */
#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif
/*
 *        the t(ype) codes of a key value     (currently only string supported)
 */
#define    tSTRING       1             /* key is a string                   */
#define    tDECIMAL      2             /* key is floating point decimal     */
#define    tINTEGER      3             /* key is signed 16 bit integer      */
#define    tLONG         4             /* key is signed 32 bit integer      */
#define    tSHORT        5             /* key size is only 8 bit integer    */
#define    tDATA         6             /* file is actually data, not keys   */
/*
 *  (rarely needed) defines for the treeprint routine
 */
#define    NOTFOUND     0
#define    PREORDER     1              /* do an pre-order tree print        */
#define    INORDER      2              /* do an in-order tree print         */
/*
 *  end of btgconst.h
 ****************************************************************************/
/*
 *  entry.h     All entry points for the BTree Library that any
 *              application should need. This file should be
 *              #included into all programs that call BTree routines
 */
extern void lioInit(void);   /* initialize the b-tree structures          */
                             /* this is called automatically now, but     */
                             /* remains here for compatability.           */

extern long lioSize(         /* how many data records/keys are in the file*/
                    int keyDes
                   );
extern long lioAvl(          /* the number of the next available data rec */
                   int keyDes
                   );
extern int lioClose(         /* close a data or index file                */
                   int keyDes
                   );
extern int lioCreat(         /* create or test for the existance of a file*/
                    char *fName,
                    int keySiz,
                    int keyTyp,
                    int dupAllowed
                   );
extern int lioFree(          /* signal that a data record is to be deleted*/
                    int keyDes,
                    long discard
                  );
extern int lioOpen(          /* open a key or data file for processing    */
                    char *fName
                  );
extern int lioRead(          /* read a record from a data file            */
                    int keyDes,
                    long pageNo,
                    char *dest
                  );
extern int lioWrite(         /* write a record to a data file             */
                    int keyDes,
                    long pageNo,
                    char *src
                   );
extern int lioFlush(         /* synchronize all files (flush buffers)     */
                    int keyDes
                   );
extern long lioInsert(       /* add a key to an index file                */
                        int keyDes,
                        char *insKey,
                        long insRcdNo
                     );
extern long lioSearch(       /* search for a key in an index file         */
                        int keyDes,
                        char *insKey
                     );
extern long lioSerNum(       /* search for the given key & record number  */
                        int keyDes,
                        char *key,
                        long num
                     );
extern long lioDelete(       /* remove a key from an index file           */
                        int keyDes,
                        char *delKey
                     );
extern long lioPrev(         /* find that key just < previous search key  */
                    int keyDes,
                    char *foundKey
                   );
extern long lioNext(         /* find that key just > previous search key  */
                    int keyDes,
                    char *foundKey
                   );
extern long lioSrPrtl(       /* find the key <= a given search value      */
                        int keyDes,
                        char *srchKey,
                        char *foundKey
                     );
extern long lioFirst(        /* find the first key in a tree              */
                        int keyDes,
                        char *fk
                    );
extern long lioLast(         /* find the last key in a tree               */
                    int keyDes,
                    char *fk
                   );
extern int lioPSize(         /* return the record length of a file        */
                    int keyDes
                   );
extern void lioGetCur(       /* save BTree currency info                  */
                        int keyDes,
                        struct currency *p
                     );
extern void lioPutCur(       /* restore currency inf                      */
                        int keyDes,
                        struct currency *p
                     );
/*
 *  end of entry.h
 */
/*----------------------- undocumented headers ---------------------- */
extern int lioDup(             /* check whether duplicates are allowed */
                    int keyDes
                 );
/***********************************************************************/
/*
 *  ientry.h    The entry points for the isam package
 */
extern int btCreate(
                    char *name,
                    int datasize,
                    int *keys,
                    int dups
                   );
extern int btOpen(
                    char *fname
                 );
extern int btClose(
                    int fd
                  );
extern int btSync(
                    int fd
                 );
/*--------------------------------------------------------------*/
extern int btAdd(
                    int fd,
                    char *data
                );
extern int btRead(
                    int fd,
                    char *data
                 );
extern int btFind(
                    int fd,
                    char *data
                 );
extern int btWrite(
                    int fd,
                    char *data
                  );
extern long btCheck(
                    int fd,
                    int inum,
                    char *key,
                    long recno
                   );
/*-------------------------------------------------------------------*/
extern int btStart(
                    int fd,
                    char *data
                  );
extern int btNext(
                    int fd,
                    char *data
                 );
extern int btPrev(
                    int fd,
                    char *data
                 );
extern int btEnd(
                    int fd,
                    char *data
                );
extern int btFPos(void);
/*-------------------------------------------------------------------*/
extern int btDelete(
                    int fd
                   );
extern long btGetCur(
                    int fd,
                    char *data
                    );
extern int btSetCur(
                    int fd,
                    long recno
                   );
extern int btSelect(
                    int fd,
                    int keyNum
                   );
/*-------------------------------------------------------------------*/
extern int btDelInx(
                    int fd,
                    int indexNum
                   );
extern int btAddInx(
                    int fd,
                    int iStart,
                    int iLen,
                    int verbose
                   );
extern long btSize(
                    int fd
                  );
extern char *btGetErr(void);
/*
 *  end of ientry.h
 *********************************************************************/
/*
 *  vlen.h
 *
 *  structure used for passing vlen paramaters back and forth
 */
struct vDetail {
        int minRec;                /* smallest record length allowed       */
        int maxRec;                /* largest record allowed               */
        long count;                /* number of records in mapper          */
        long mDel;                 /* number of deleted mapper entries     */
        long stDel;                /* number of deleted bytes in the store */
        long stLast;               /* last byte in use in the store        */
    };

/*
 *  that's all, folks!
 */
extern  int vlCreat(char *fName);
extern  int vlOpen(char *fName);
extern  int vlClose(int fd);
extern  int vlFlush(int fd);
extern  long vlAvl(int fd);
extern  int vlWrite(int fd,long mapRp,char *data,int len);
extern  int vlFree(int fd,long mapRp);
extern  int vlRead(int fd,long mapRp,char *data);
extern  char *vlGetErr(void );
extern  int vlStats(int fd,struct vDetail *p);
/*------------------------------------------------------------------------*/
/*
 *  keyBuild.h  Useful keybuilding constatnts
 */
#define ASCENDING   1   /* ascending order key                      */
#define DESCENDING  -1  /* descending order key                     */
#define EOK         -99 /* end of keys                              */
/*------------------------------------------------------------------*/
extern  int keyDef(
                    int kd,
                    char *data,
                    char *keydest,
                    int *rules
                  );
extern  int keyBuild(
                    int kd
                    );
/********************************************************************
 *   entry points for the memory functions
 */
extern  int exist(char *fName);
extern  void ltoc(long j,  char *buff);
extern  long ctol(unsigned char *buff);
extern  int rmsign(unsigned char c);
/********************************************************************//*
 *  screen.h    The entry points for the screen driver module
 */
extern void termInit(void);   /* initialize a given terminal  */
extern void termExit(void);   /* de-initialize a terminal     */
extern void curPos(           /* move the cursor absolute     */
                    int row,
                    int col
                  );
extern void scrClr(void);     /* clear the screen             */
extern void eraEol(           /* erase to end of line         */
                    int row,
                    int col
                  );
extern void eraEop(           /* erase from the cursor to the end of screen */
                    int row,
                    int col
                  );
extern int getkey(void);
/*
 *  end of screen.h
 */
/*
 * ---------------------------- INPUT.h ----------------------------
 */
extern  void input(
                    char *str,
                    int maxlen
                  );
static  void eraChr(void );
static  void pstr(
                    char *str
                 );

/*  - , ᯮ,  祣  */
