/*
 *   SoftFocus BTree Library: v2.0
 *
 *      Copyright 1984, 1985 by Softfocus
 *                              1343 Stanbury Dr.,
 *                              Oakville, ONT.
 *                              Canada, L6L 2J5
 *                              (416) 825 - 0903
 *
 *   Written by L.Karnis from Dec., 83 to July 84
 *
 *   Remember your license states that you can not redistribute the
 *   source, manual or make public the object code with the entry
 *   points. You can use this software in any application, flog that
 *   and you don't owe us a penny. Please keep us informed if you
 *   find any bugs or make any significant enhancements.
 *
 *  Version 2.0     Released 85.05.01
 *
 *  The significant difference between this and earlier
 *  versions is the support of duplicate keys. Some algorithms have been
 *  streamlined and some (deeply buried) bugs exterminated.
 *
 *  We have tested this package with 60,000 records, key sizes up to
 *  80 bytes long and data records up to 10K bytes long.
 *
 *  jon simkins, 85.04.22
 *
 *  Date        Revision History
 *  ----        -------- -------
 *
 *  85.04.28    Fixed up functions that copied new key values to
 *              NOT return the duplicate bytes.
 *              mem2cpy no longer used and kkSize() added. j.s.
 *
 *  85.05.04    Fixed lioAvl() and lioFree() to release malloc()ed
 *              space in those routines in case of btRead() or
 *              btWrite() failure. js.
 *
 *  85.05.22    (How did we miss this one??)
 *              Fixed bug in lioSearch() that would cause search
 *              failure of duplicate keys after the first one entered
 *              had been deleted. Changed ndxCmp1() to take an explicit
 *              byte count instead of reading index[]. js
 *
 *  85.05.23    Optimization to btSearch() & btDelete() to factor out
 *              unnecessary ndxCmp() calls. js (suggested by David
 *              Caulfield)
 *
 *  85.06.07    Changed ctol() and ltoc() to use shift instructions
 *              instead of long multiplies. Old routines commented
 *              out (just in case). js.
 *
 *  85.06.17    Added correct declaration for memset() and memcpy();
 *              this solved problems under C86 big model. Also added
 *              cast to second argument in memset(). js.
 *
 *  85.07.30    Fixed bug in lioOpen() that could close stdout if the
 *              first file open operation failed. thank you Steve. js.
 *
 *  85.09.19    Changed lioSerNum() to conform with the manual description.
 *              js.
 *
 *  85.09.30    Fixed lioSerNum() which would occaisionally cause
 *              spurious 'no key for data record' in ISAM.
 *              Linted and cleaned up formatting using Wizard's C
 *              comprehensive diagnostics! js.
 *
 *  85.10.07    Fixed lioSerNum() to work correctly (added kkSize()).
 *              js. I hate lioSerNum().
 *
 *  85.10.11    Fixed a bug in ndxCmp(), where it was giving the first
 *              duplicate in a tree the same duplicate number as the
 *              original. larry karnis (yes I'm still alive!)
 *
 *  85.10.14    Added conditional compilation of rmsign() routines in
 *              ctol() for compilers that sign extend chars. js.
 *
 *  85.10.18    Added (int) casts in inxCmp1() for DeSmet. Added char
 *              sign extension logic into inxCmp1() (probably paranoid
 *              code!) js.
 *
 *  85.12.20    lioOpen now copies the name the file was opened with into
 *              the header block. Could cause problems if the file was
 *              renamed and lioFlush()ed. Thanks Dave! Merry Christmas.js.
 *
 *  86.01.22    Cut ltoc() and ctol() out of BTree and put them in system.c.
 *              Fixed (tiny) problem with lioDelete() that could fail in
 *              certain cases when called with an explicit key. js.
 *
 *  86.03.04    lioGetCur() and lioPutCur() improved and tested. Still
 *              formally undocumented, but now working. js.
 *              Program carefully reformatted. js.
 *              uTag union pulled out; only struct detail now used for
 *              header read/writes. js.
 *
 *  86.03.20    Added a lru cache algorithm for tree pages. Speed increase
 *              of 10-30% depending on operation. js.
 *              Fixed bug in search partial that could find next largest
 *              key even when given an exact match. thanks, bob. js.
 *
 *  86.05.12    Fixed up some internal declarations that gave DeSmet users
 *              some grief. js.
 *
 *  86.06.06    Cleaneded up some bad casts. js.
 *
 *  86.06.25    Added explicit binary open() after creat() for MicroSoft.
 *              Who started this ASCII/binary file mode stuff anyway? js.
 *
 *  86.09.11    Insert speed bumped up about 25%; added treeShuffle and test
 *              in insert. Thanks to Ernie for lighting the fires!
 *              js (with some guidance from larry).
 *
 *  86.09.22    Fixed ndxCmp() call that could mess up sequential reads if
 *              there was no trailing NULL on the key and fast ltoc() routines
 *              were enabled. js.
 *
 *  86.11.18    Cleaned up cache flush logic after key page deletes. Thanks to
 *              Scott Taggart for doing most of the work, and Jim Benson for
 *              providing the test case. js.
 *
 *  99.04.26    Added new key types: short, int, long, double. [Y.Krut]
 *               :
 *                  o  : -       ,
 *                                 1  (short, ushort),
 *                                 2  (integer, uinteger),
 *                                 3  (triple) 
 *                                 4  (long, ulong);
 *                  o            -    ,
 *                                 4  (double);
 *                               -  , 
 *                                  '\0';
 *                  o
 *                  o
 */
#include <stdio.h>  /* compiler manual, also common defines                 */
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "include/btgconst.h"   /* all useful B-Tree global constants               */
#include "include/entry.h"
#include "include/system.h"     /* entry points for system dependent functions      */

/*
#pragma intrinsic       memcpy
#pragma intrinsic       memset
*/

/*
 *    general useful constants, but only used in this file.
 */
#define    MAXKEYS  2*MINKEYS      /* b-tree page size                      */
#define    loop     while (TRUE){
#define    endLoop  }
#define    LERROR   -1L
#define    NUMFIELDS    10         /* max number of fields in key - NOT USED */
#define    FILEXIST NULL           /* access mode for checking file pres     */
#define    NOCREAT  NULL           /* creat returns 0 on error in unix       */
#define    NOPEN    ERROR          /* open returns -1 on error in unix       */
/*
 *   each page entry has the following attributes
 */
struct ITEM {       /* definition of a list of entries            */
   char *srchVal;   /* each entry has a key                       */
   long spare;      /* the data file record number                */
   long child;      /* a pointer to entries > key and < key + 1   */
};
/*
 *    the B-tree node structure
 */
struct PAGE {
   short       nodSiz;            /* number of items in this node */
   long        lftSon;            /* pointer to leftmost subpage  */
   struct ITEM nodElmt[MAXKEYS];  /* list of keys & page pointers */
};
/*
 *    this is what a file header block looks like.
 */
LOCAL struct {
        char  lioName[NAMESIZE]; /* name of the file                       */
        short order;             /* the order of the b-tree                */
        short kType;             /* keys of different types supported?     */
        short kSize;             /* size of the key field (in bytes)       */
        long  freeNum;           /* log-rec # for first entry free l-list  */
        int   pageSize;          /* the number of bytes in each page rec.  */
        long  rootNum;           /* logical record number of the root page */
        long  numKeys;           /* how many keys currently allocated      */
        long  numRcds;           /* how many pages currently allocated     */
        short inUse;             /* set on file open, reset on close       */
        struct {                 /* NOT USED AT THIS TIME                  */
            short fieldNo;
            short startCh;
            short endCh;
        } keyForm[NUMFIELDS];
        short dupOk;             /* are duplicates allowed on this index?   */
        long  dupVal;            /* unique id assigned to a duplicate insert*/
    } detail;

LOCAL int initialized = (int) NULL; /* used to initialize the internal tables */
/*----------------------------------------------------------------------------
 *  internal (static) function declarations
 */
static int      kkSize(int btDes);
static int      kkType(int btDes);
static char     *rdCache(int keyDes,long pageNo);
static char     *wrtCache(int keyDes,long pageNo);
static int      flCache(void );
static int      lioRdKey(int keyDes,long pageNo,struct PAGE *dest);
static int      lioWrtKey(int keyDes,long pageNo,struct PAGE *src);
static void     itemcp(int keyDes,struct ITEM *dest,struct ITEM *src);
static int      markFile(int keyDes);
static long     makNod(int keyDes,struct PAGE *ptr);
static void     nodFree(void );
static int      btSearch(int keyDes,struct PAGE *subTrRoot,struct ITEM *holdItem,long subTrPtr);
static int      btInsert(int keyDes,struct PAGE *subTrRoot,struct ITEM *thisItem,int keyPositn,struct ITEM *holdItem);
static int      btDelete(int keyDes,struct PAGE *subTrRoot,long subTrPtr);
static int      underflow(int keyDes,struct PAGE *ancestor,long subTrRoot,int srchPosn);
static int      getRep(int keyDes,long curNode,struct PAGE *subTrRoot,int k);
static int      prolog(int keyDes);
static long     trInsert(int keyDes,char *insKey,long insRcdNo);
static int      btFirst(int keyDes,long rootNod,char *fk);
static int      btLast(int keyDes,long rootNod,char *fk);
static int      ndxCmp(int keyDes,char *left,char *right);
static char     *nodAlloc(int size);
/*---------------------------------------------------------------------------
 *  detail for open files lives here
 */
LOCAL struct indexTag  { /* global variables required to maintain the key    */
   int   keySize;        /* files on a data file.                            */
   int   keyType;
   int   keyOrder;
   int   keyPgSz;
   int   fd;

   short trInUse;
   long  numRecs;                 /* how many pages currently allocated   */
   long  nKeys;                   /* how many keys r currently allocated  */
   long  nextAvl;                 /* next logical record on the free list */
   long  treeRoot;                /* logical record number of root node   */
   short stackPtr;                /* record # path to last page searched  */
   long  stack[MAXTREEHEIGHT];    /* remember path for sequential searchs */
   char  matchName[MAXKEYSIZE+1]; /* last key matched by sequential search*/
   long  lastRcd;                 /* record number of last record found   */
   short hasDup;                  /* duplicates allowed on this index?    */
   long  dupNum;                  /* next available duplicate number      */
   char  fName[NAMESIZE];         /* name the file was opened with        */
} Index[NUMINDEXS];

LOCAL short       hasOver;         /* has the tree overflowed                */
LOCAL short       treeShuffle;     /* flag to indicate that tree has changed */
LOCAL short       amInserting;     /* true while actually inserting a key    */
LOCAL struct PAGE rootPg;          /* workspace for a new root page          */
LOCAL struct ITEM currItem;        /* temporary item for inserting a key     */
LOCAL char currBufr[MAXKEYSIZE+1]; /* more temporary key space               */
LOCAL char keyValue[MAXKEYSIZE+1]; /* current key being operated on          */
LOCAL short       adjFlag;         /* true if something is being passed back up*/
LOCAL long        btResult;        /* status word for success or failure of an op'n    */
LOCAL long        keyRcdNo;        /* data file record number for what is being opd on */
LOCAL struct {
    short pgInUse;
    char *pgPtr;
} allocPage[MAXTREEHEIGHT*2];
/*-----------------------------------------------------------------------
 *  cache details
 *
 *  Our tests here (on an 8mhz 286 MS-DOS machine) indicate that 5 is
 *  a near-optimum number of slots; you may wish to experiment with your
 *  own flavour of hardware.
 *
 *  Setting MAXCACHE to 1 effectively disables all cache operations.
 */
#define MAXCACHE 5
LOCAL unsigned lru = 1;
LOCAL struct {
    int      cfd;
    long     cRecNum;
    int      dirty;
    unsigned lru;
    char     temp[ (MAXKEYS*(MAXKEYSIZE+6)) + 3 + sizeof(char)];
} cache[MAXCACHE];
/* ------------------------------------------------------------------------ */
/*
 *  initialize the in memory tables. Remains an entry point for compatibility.
 */
ENTRY void lioInit(void)
{
   if(initialized)
       return;
   memset((char *)Index,     (char) NULL, sizeof(Index));
   memset((char *)allocPage, (char) NULL, sizeof(allocPage));
   initialized = TRUE;
   currItem.srchVal = currBufr;         /* set up some temp. key space */
}
/*
 *  The following two routines are used to save and restore the BTree
 *  currency. The structure 'currency' is defined in btgconst.h.
 *
 *  copy out the current currency values
 */
ENTRY void lioGetCur(keyDes, p)
int             keyDes;
struct currency *p;
{
    if(!initialized)
        lioInit();
    p->sPtr = Index[keyDes].stackPtr;
    memcpy((char *) p->stk, (char *) Index[keyDes].stack,
             sizeof(long)*MAXTREEHEIGHT );
    memcpy(p->last, Index[keyDes].matchName, Index[keyDes].keySize);
}
/*
 *   replace the currency information
 */
ENTRY void lioPutCur(keyDes, p)
int             keyDes;
struct currency *p;
{
    if(!initialized)
        lioInit();
    Index[keyDes].stackPtr = p->sPtr;
    memcpy((char *) Index[keyDes].stack, (char *) p->stk,
             sizeof(long)*MAXTREEHEIGHT );
    memcpy(Index[keyDes].matchName, p->last, Index[keyDes].keySize);
}
/*
 *    return the number of keys/records for a data or index file
 */
ENTRY long lioSize(keyDes)
register int keyDes;
{
    if(!initialized)
        lioInit();
    return(Index[keyDes].keyType == tDATA) ? Index[keyDes].numRecs :
                         Index[keyDes].nKeys;
}
/*
 *   return the size of a 'unit' of storage.
 */
ENTRY int lioPSize(keyDes)
register int keyDes;
{
    if(!initialized)
        lioInit();
    return Index[keyDes].keyPgSz;
}
/*
 *   return a boolean that says whether duplicates are allowed
 */
ENTRY int lioDup(keyDes)
int keyDes;
{
    if(!initialized)
        lioInit();
   return Index[keyDes].hasDup;
}
/*
 *   get the logical record number of the next free record
 */
ENTRY long lioAvl(keyDes)
register int keyDes;
{
    register long result;
    char *ptr;

    if(!initialized)
        lioInit();
    Index[keyDes].numRecs++;
    /*
     *  set treeShuffle (only meaningful within lioInsert())
     */
    treeShuffle = TRUE;
    /*
     *  see if there's anythin on the free list
     */
    if(Index[keyDes].nextAvl != (int) NULL){
        result = Index[keyDes].nextAvl;
		if((ptr = malloc(Index[keyDes].keyPgSz)) == NULL)
            return LERROR;
        if(lioRead(keyDes, result, ptr) == (int) NULL){
            free(ptr);
            return LERROR;
        }
        memcpy((char *) &Index[keyDes].nextAvl, ptr, sizeof(long));
        free(ptr);
        return result;
    }
    return Index[keyDes].numRecs;
}
/*
 *  close an index file
 */
ENTRY int lioClose(keyDes)
register int keyDes;
{
    if(!initialized)
        lioInit();
    if(lioFlush(keyDes))
        return ERROR;
    if(close(Index[keyDes].fd))
        return ERROR;
    /*
     *  NULL out the files' old entry in the control table
     */
    Index[keyDes].keySize  = (int)NULL;
    Index[keyDes].keyType  = (int)NULL;
    Index[keyDes].keyOrder = (int)NULL;
    Index[keyDes].keyPgSz  = (int)NULL;
    Index[keyDes].numRecs  = LNULL;
    Index[keyDes].nKeys    = LNULL;
    Index[keyDes].fd       = (int)NULL;
    return( (int)NULL );
}
/*
 *  flush an index or data file
 */
ENTRY int lioFlush(keyDes)
int keyDes;
{
    /*
     *  flush the cache
     */
    if(Index[keyDes].keyType != tDATA)
        if(flCache() == ERROR)
            return ERROR;
    /*
     *   when a file is released, reset the inUse flag to indicate that
     *   file operations were terminated normally.
     */
    if(!initialized)
        lioInit();

    if(lseek(Index[keyDes].fd, LNULL, (int) NULL) == LERROR)
        return ERROR;
    if(read(Index[keyDes].fd,(char *) &detail, sizeof(detail))
                                                          != sizeof(detail))
        return ERROR;
    /*
     *  update the file header block values
     */
    memset(detail.lioName, (char) NULL,         NAMESIZE);
    memcpy(detail.lioName, Index[keyDes].fName, NAMESIZE);
    detail.freeNum = Index[keyDes].nextAvl;
    detail.numRcds = Index[keyDes].numRecs;
    detail.numKeys = Index[keyDes].nKeys;
    detail.rootNum = Index[keyDes].treeRoot;
    detail.dupVal  = Index[keyDes].dupNum;
    detail.inUse   = Index[keyDes].trInUse = FALSE;

    if(lseek(Index[keyDes].fd, LNULL, (int) NULL) == LERROR)
        return ERROR;
    if(write(Index[keyDes].fd, (char*)&detail, sizeof(detail))
                                                     != sizeof(detail))
        return ERROR;
    /*
     *  only safe GENERIC way to ensure file is on disk using MS-DOS; why
     *  isn't sync() in the operating system somewhere?
     */
    close(Index[keyDes].fd);
    Index[keyDes].fd = open(detail.lioName, RWMODE);

    return (int) NULL;
}
/*
 *    Initialize a new Index file.
 */
ENTRY int lioCreat(fName, keySiz, keyTyp, dupAllowed)
char *fName;
int keySiz, keyTyp, dupAllowed;
{
    register int fd;

    if(!initialized)
        lioInit();
    if(keySiz > MAXKEYSIZE-3 && keyTyp != tDATA && dupAllowed)
        return ERROR;
    memset((char *) &detail, (char) NULL, sizeof(detail));
    memcpy(detail.lioName, fName, NAMESIZE);
    if(exist(detail.lioName))
        return ERROR;
    /*
     *  actually create the file here. Explicit re-open is to ensure
     *  binary rw mode.
     */
    if((fd = creat(detail.lioName, CREATMODE)) == (int) NOCREAT)
        return ERROR;
    if(close(fd) == ERROR)
        return ERROR;
    if((fd = open(detail.lioName, RWMODE)) == NOPEN)
        return ERROR;
    /*
     *  build and write file header
     */
    detail.order = MINKEYS;      /* b-tree order (for safe keeping) */
    if(dupAllowed && keyTyp != tDATA) keySiz += 3;
    detail.kSize = keySiz;
    detail.kType = keyTyp;       /* either tDATA or tSTRING         */
    detail.dupOk = dupAllowed;   /* remember if duplicates are ok   */
    /*
     * calculate the number of bytes required for each page
     * of the key tree
     */
    detail.pageSize = (keyTyp == tDATA) ? keySiz :
        (keySiz+2*(sizeof(long)-1))*MAXKEYS+(sizeof(long)-1)+sizeof(short);
    /*
     *  write the file header
     */
    if(write(fd, (char *) &detail, sizeof(detail)) != sizeof(detail))
        return ERROR;
    if(close(fd))
        return ERROR;
   return (int) NULL;
}
/*
 *  return an allocated logical record number to the free list
 */
ENTRY int lioFree(keyDes, discard)
int keyDes;
long discard;
{
    char *ptr;

    if(!initialized)
        lioInit();
	if((ptr = malloc(Index[keyDes].keyPgSz)) == NULL)
        return ERROR;
    memset(ptr, (char) NULL,                   Index[keyDes].keyPgSz);
    memcpy(ptr, (char*)&Index[keyDes].nextAvl, sizeof(long));
    /*
     *  clean out the cache if this is a key file
     */
    if(Index[keyDes].keyType != tDATA)
        flCache();

    if(lioWrite(keyDes, discard, ptr) == (int) NULL){
        free(ptr);
        return ERROR;
    }
    Index[keyDes].numRecs--;
    Index[keyDes].nextAvl = discard;
    free(ptr);

    return (int) NULL;
}
/*
 *  lioOpen - open an existing key or data file.
 */
ENTRY int lioOpen(fName)
char *fName;
{
    register int i, keyDes;

    if(!initialized)
        lioInit();
    /*
     *  find the first empty file control block
     */
    memset( (char *)&detail, (char) NULL, sizeof(detail));
    for(i = 0; i < NUMINDEXS; i++)
        if(Index[i].fd == (int) NULL)
            break;
    if((keyDes = i) == NUMINDEXS)
        return ERROR;
    /*
     *  open the file and read the header
     */
    if((Index[keyDes].fd = open(fName, RWMODE)) == NOPEN){
        Index[keyDes].fd = (int) NULL;
        return ERROR;
    }
    if(read(Index[keyDes].fd, (char *) &detail, sizeof(detail))
                                                    != sizeof(detail))
        return ERROR;
    /*
     *  check that the file is uncorrupted
     */
    if(detail.inUse){
        (void) close(Index[keyDes].fd);
        Index[keyDes].fd = (int) NULL;
        return ERROR;
    }
    /*
     *  set up the in-memory table
     */
    memcpy(Index[keyDes].fName, fName, NAMESIZE);
    Index[keyDes].nextAvl  = detail.freeNum;
    Index[keyDes].numRecs  = detail.numRcds;
    Index[keyDes].nKeys    = detail.numKeys;
    Index[keyDes].keySize  = detail.kSize;
    Index[keyDes].keyType  = detail.kType;
    Index[keyDes].keyPgSz  = detail.pageSize;
    Index[keyDes].treeRoot = detail.rootNum;
    Index[keyDes].hasDup   = detail.dupOk;
    Index[keyDes].dupNum   = detail.dupVal;

    return keyDes;
}
/*
 *  get a logical page from the file fd
 */
ENTRY int lioRead(keyDes, pageNo, dest)
register int keyDes;
register long pageNo;
char *dest;
{
    long offset;

    if(!initialized)
        lioInit();

    offset = sizeof(detail) + (pageNo-1)*Index[keyDes].keyPgSz;
    if(lseek(Index[keyDes].fd, offset, (int) NULL) == LERROR)
        return (int) NULL;
    if(read(Index[keyDes].fd,dest, Index[keyDes].keyPgSz) !=
                                                    Index[keyDes].keyPgSz)
        return (int) NULL;

    return Index[keyDes].keyPgSz;
}
/*
 *  put a logical page to the file fd
 */
ENTRY int lioWrite(keyDes, pageNo, src)
register int keyDes;
register long pageNo;
char *src;
{
    long offset;

    if(!initialized)
        lioInit();

    if(Index[keyDes].trInUse == FALSE){
        Index[keyDes].trInUse = TRUE;
        if(markFile(keyDes))
            return (int) NULL;
    }
    offset = sizeof(detail) + (pageNo - 1)*Index[keyDes].keyPgSz;
    if(lseek(Index[keyDes].fd, offset, (int) NULL) == LERROR)
        return (int) NULL;
    if(write(Index[keyDes].fd, src, Index[keyDes].keyPgSz) !=
                                                    Index[keyDes].keyPgSz)
        return (int) NULL;
    return Index[keyDes].keyPgSz;
}
/*
 *  kkSize  Return the size of the actual index keysize MINUS any
 *          duplicate key bytes.
 */
LOCAL int kkSize(btDes)
int btDes;
{
    return( (Index[btDes].hasDup) ? Index[btDes].keySize - 3 :
                                    Index[btDes].keySize      );
}
/*
 *  kkType  Return the type of the index key.
 */
LOCAL int kkType(btDes)
int btDes;
{
    return( Index[btDes].keyType );
}
/*
 *  rdCache     Return pointer to the requested page. Update
 *              the cache if necessary.
 */
LOCAL char *rdCache(keyDes, pageNo)
int  keyDes;
long pageNo;
{
    register j;
    unsigned last, slot;
    /*
     *  see if the page in is the cache already
     */
    for(j=0, last=lru, slot=0; j<MAXCACHE; j++){
        if((cache[j].cRecNum == pageNo) && (cache[j].cfd == keyDes) ){
            cache[j].lru = lru++;
            return cache[j].temp;
        }
        if(last > cache[j].lru){
            last = cache[j].lru;
            slot = j;
        }
        if(cache[slot].cRecNum == 0L)
            break;
    }
    /*
     *  write dirty slot
     */
    if(cache[slot].dirty)
        if(lioWrite(cache[slot].cfd, cache[slot].cRecNum, cache[slot].temp)==
            (int) NULL)
            return (char *) NULL;
    /*
     *  read the given record and update the cache
     */
    if(lioRead(keyDes, pageNo, cache[slot].temp) == (int) NULL)
        return (char *) NULL;

    cache[slot].cRecNum = pageNo;
    cache[slot].cfd     = keyDes;
    cache[slot].dirty   = FALSE;
    cache[slot].lru     = lru++;

    return( cache[slot].temp );
}
/*
 *  wrtCache        Return a pointer to where the given page should be written.
 *                  Flush the cache as necessary.
 */
LOCAL char *wrtCache(keyDes, pageNo)
int keyDes;
long pageNo;
{
    register j;
    unsigned slot, last;
    /*
     *  see if the page in is the cache already
     */
    for(j=0, last=lru, slot=0; j<MAXCACHE; j++){
        if( (cache[j].cRecNum == pageNo) && (cache[j].cfd == keyDes) ){
            slot = j;
            cache[slot].dirty = FALSE;
            break;
        }
        if( (last > cache[j].lru) ){
            slot = j;
            last = cache[j].lru;
        }
        if(cache[slot].cRecNum == 0L)
            break;
    }
    /*
     *  write slot (if necessary) and update cache entries
     */
    if(cache[slot].dirty)
        if(lioWrite(cache[slot].cfd, cache[slot].cRecNum, cache[slot].temp)==(int) NULL)
            return (char *) NULL;

    cache[slot].lru = lru++;
    cache[slot].cRecNum = pageNo;
    cache[slot].cfd = keyDes;
    cache[slot].dirty = TRUE;

    return cache[slot].temp;
}
/*
 *  flCache     Flush all dirty buffers to disk. Reset lru to the
 *              beginning. Clear the cache.
 */
LOCAL int flCache(void)
{
    register j;

    for(j=0; j<MAXCACHE; j++){
        if(cache[j].cRecNum == 0L)
            break;
        if(cache[j].dirty)
            if(lioWrite(cache[j].cfd, cache[j].cRecNum, cache[j].temp) ==
                                                               (int)NULL)
                return ERROR;
    }
    memset((char *) cache, '\0', sizeof(cache));
    lru = 1;

    return OK;
}
/*
 *  lioReadKey  read a page from the disk, get space for all active
 *              keys and install pointers to them in the node.
 */
LOCAL int lioRdKey(keyDes, pageNo, dest)
int         keyDes;             /* index into key descriptor table  */
long        pageNo;             /* logical page to read             */
struct PAGE *dest;              /* where to put the adjusted node   */
{
    char *dSpace;
    char *buff;
    register int i, len;

    if((buff = rdCache(keyDes, pageNo)) == (char *) NULL)
        return ERROR;

    len = Index[keyDes].keySize;

    dSpace = nodAlloc(MAXKEYS*(len+1));
    dest->nodSiz = *buff++;
    dest->lftSon = ctol(buff);
    buff += 3;

    for(i = (int) NULL; i < MAXKEYS; i++){
        dest->nodElmt[i].srchVal = dSpace + i*(len+1);
        if(i < dest->nodSiz){
            memcpy(dest->nodElmt[i].srchVal, buff, len);
            buff += len;
            dest->nodElmt[i].child = ctol(buff);  /* get sub-tree pointer */
            buff += 3;
            dest->nodElmt[i].spare = ctol(buff);  /* & the data record #  */
            buff += 3;
        }
    }
    return (int) NULL;
}
/*
 *  lioWriteKey     write a page to the disk, rebuild the node before
 *                  writing.
 */
LOCAL int lioWrtKey(keyDes, pageNo, src)
int         keyDes;             /* index into key descriptor table */
long        pageNo;             /* logical page to read            */
struct PAGE *src;               /* where to put the adjusted node  */
{
    char *buff;
    int i, len;

    if( (buff = wrtCache(keyDes, pageNo)) == (char *) NULL)
        return ERROR;

    len = Index[keyDes].keySize;
    *buff++ = (char)(src->nodSiz);      /* added "(char)". vs*/
    ltoc(src->lftSon, buff);
    buff += 3;

    for(i = 0; i < MAXKEYS; i++)
        if(i < src->nodSiz){
            memcpy(buff, src->nodElmt[i].srchVal, len);
            buff += len;
            ltoc(src->nodElmt[i].child, buff);
            buff += 3;
            ltoc(src->nodElmt[i].spare, buff);
            buff += 3;
        }
    return (int) NULL;
}
/*
 *  item copy - copy one node item to another
 */
LOCAL void itemcp(keyDes, dest, src)
register struct ITEM    *dest, *src;
register int            keyDes;
{
    memcpy(dest->srchVal, src->srchVal, Index[keyDes].keySize);
    dest->child = src->child;
    dest->spare = src->spare;
}
/*
 *    markFile - mark a file as being modified but not flushed.
 */
LOCAL int markFile(keyDes)
register int keyDes;
{

    if(lseek(Index[keyDes].fd, LNULL, (int) NULL) == LERROR)
        return ERROR;
    if(read(Index[keyDes].fd, (char *) &detail, sizeof(detail))
                                                    != sizeof(detail))
        return ERROR;
    detail.inUse = TRUE;
    if(lseek(Index[keyDes].fd, LNULL, (int) NULL) == LERROR)
        return ERROR;
    if(write(Index[keyDes].fd, (char *) &detail, sizeof(detail)) !=
                                                            sizeof(detail))
        return ERROR;

    return (int) NULL;
}
/*
 *    node initializer - null out a new node
 */
LOCAL long makNod(keyDes, ptr)
register int keyDes;
struct PAGE *ptr;
{
    register int i;
    char         *pgPtr;

    pgPtr = nodAlloc(MAXKEYS*(Index[keyDes].keySize+1));
    ptr->nodSiz = (int) NULL;       /* no entries in this page         */
    ptr->lftSon = LNULL;            /* leftmost leaf not allocated     */

    for(i=0; i < MAXKEYS; i++){
        ptr->nodElmt[i].srchVal = pgPtr + i*(Index[keyDes].keySize+1);
        ptr->nodElmt[i].child = LNULL;  /* no pages below this one    */
        ptr->nodElmt[i].spare = LNULL;  /* no occurrances of item     */
    }
    return lioAvl(keyDes);         /* grab the next entry off free list  */
}
/*
 *    dynamically allocate space for a newly created or just read node.
 */
LOCAL char *nodAlloc(size)
int size;             /* number of bytes to allocate        */
{
    register int i;
    char         *ptr;

	if((ptr = malloc(size)) == NULL){
        puts("nodAlloc: can't allocate any more space\n");
        exit( (int) NULL);
    }

    memset(ptr, (char) NULL, size);
    for(i=0; i<2*MAXTREEHEIGHT; i++)
        if(allocPage[i].pgInUse==FALSE)
            break;

    if(i == 2*MAXTREEHEIGHT){
        puts("nodAlloc: no page stack space left\n");
        exit(ERROR);
    }
    allocPage[i].pgInUse = TRUE;        /* deleted in one shot    */
    allocPage[i].pgPtr   = ptr;         /* at end of the command  */
    return( ptr );
}
/*
 *    free the pages allocated by the above function
 */
LOCAL void nodFree()
{
    register int i;

    for(i = 0; i < 2*MAXTREEHEIGHT; i++){
        if(allocPage[i].pgInUse)
            free(allocPage[i].pgPtr);
        allocPage[i].pgInUse = FALSE;
        allocPage[i].pgPtr   = NULL;
    }
}
/************************** End of B-Tree Clerical F'ns **********************/
/*
 *    btSearch for key keyValue on the b-tree with root subTrRoot.
 */
LOCAL int btSearch(keyDes, subTrRoot, holdItem, subTrPtr)
int         keyDes;
struct PAGE *subTrRoot;
struct ITEM *holdItem;
long        subTrPtr;
{
    register int  k, leftSide, rightBnd;
    short         stackVal;
    register long son;
    char          buffer[MAXKEYSIZE+1];
    struct ITEM   currItem;
    struct PAGE   temp;
    int           tCmp;

    currItem.srchVal = buffer;

    if(subTrPtr == (int) NULL){    /* prepare an item to be inserted   */
        adjFlag = TRUE;
        memcpy(Index[keyDes].matchName, keyValue,
                                          Index[keyDes].keySize);
        memcpy(holdItem->srchVal, keyValue, Index[keyDes].keySize);
        holdItem->child = LNULL;
        holdItem->spare = keyRcdNo;
        if(amInserting){
            Index[keyDes].nKeys++;
            btResult=keyRcdNo;
        }
    }else{
        leftSide = (int) NULL;
        rightBnd = subTrRoot->nodSiz - 1;
        do {
            k    = (leftSide + rightBnd) / 2;
            tCmp = ndxCmp(keyDes, keyValue, subTrRoot->nodElmt[k].srchVal);
            if(tCmp <= (int) NULL)
                rightBnd = k - 1;
            if(tCmp >= (int) NULL)
                leftSide = k + 1;
        } while(rightBnd >= leftSide);

        if(tCmp == (int) NULL){
            btResult = subTrRoot->nodElmt[k].spare;
            memcpy(Index[keyDes].matchName, keyValue, Index[keyDes].keySize);
            adjFlag  = FALSE;
        }else{
            son = (ndxCmp(keyDes,keyValue,subTrRoot->nodElmt[0].srchVal)
                                                                < (int) NULL)
                   ? subTrRoot->lftSon : subTrRoot->nodElmt[rightBnd].child;
            if(son){
                stackVal                      = Index[keyDes].stackPtr;
                Index[keyDes].stack[stackVal] = son;
                Index[keyDes].stackPtr++;
                if(lioRdKey(keyDes, son, &temp))
                    return ERROR;
            }
            if(btSearch(keyDes, &temp, &currItem, son))
                return ERROR;

            if(adjFlag && amInserting){
                if(btInsert(keyDes, subTrRoot, &currItem, rightBnd, holdItem))
                    return ERROR;
                if(lioWrtKey(keyDes, subTrPtr, subTrRoot))
                    return ERROR;
            }
            if(!amInserting)
                adjFlag = FALSE;
        }
    }
    return( (int)NULL );
}
/*
 *    insert the value keyValue into the b-tree
 */
LOCAL int btInsert(keyDes, subTrRoot, thisItem, keyPositn, holdItem)
int         keyDes,
            keyPositn;
struct PAGE *subTrRoot;
struct ITEM *thisItem,
            *holdItem;
{
    long          makNod();
    register long localPg;
    register      int i;
    struct PAGE   splitPg;

    if(subTrRoot->nodSiz < MAXKEYS){
        subTrRoot->nodSiz++;
        adjFlag = FALSE;
        for(i = subTrRoot->nodSiz-1; i >= keyPositn + 2; i--)
            itemcp(keyDes, &subTrRoot->nodElmt[i],
                                            &subTrRoot->nodElmt[i-1]);
        itemcp(keyDes, &subTrRoot->nodElmt[keyPositn + 1], thisItem);
    } else {
        hasOver = TRUE;
        if((localPg = makNod(keyDes, &splitPg)) == LERROR)
            return ERROR;
        if(keyPositn < MINKEYS){
            if(keyPositn == MINKEYS-1)
                itemcp(keyDes, holdItem, thisItem);
            else {
                itemcp(keyDes, holdItem, &subTrRoot->nodElmt[MINKEYS-1]);
                for(i = MINKEYS-1; i >= keyPositn + 2; i--)
                itemcp(keyDes, &subTrRoot->nodElmt[i],
                                                &subTrRoot->nodElmt[i-1]);
                itemcp(keyDes, &subTrRoot->nodElmt[keyPositn + 1], thisItem);
            }
            for(i = 0; i < MINKEYS; i++)
                itemcp(keyDes,&splitPg.nodElmt[i],
                                                &subTrRoot->nodElmt[i+MINKEYS]);
        } else {
            keyPositn -= MINKEYS;
            itemcp(keyDes, holdItem, &subTrRoot->nodElmt[MINKEYS]);
            for(i = 0; i < keyPositn; i++)
                itemcp(keyDes,&splitPg.nodElmt[i],
                                            &subTrRoot->nodElmt[i+MINKEYS+1]);
            itemcp(keyDes, &splitPg.nodElmt[keyPositn], thisItem);
            for(i = keyPositn + 1; i < MINKEYS; i++)
                itemcp(keyDes, &splitPg.nodElmt[i], 
                                            &subTrRoot->nodElmt[i+MINKEYS]);
        }
        subTrRoot->nodSiz = splitPg.nodSiz = MINKEYS;
        splitPg.lftSon = holdItem->child;
        holdItem->child = localPg;
        if(lioWrtKey(keyDes, localPg, &splitPg))
            return ERROR;
    }
    return (int) NULL;
}
/*
 *   delete     search for and delete the key keyValue in a b-tree
 *              with root subTrRoot.
 */
LOCAL int btDelete(keyDes, subTrRoot, subTrPtr)
int keyDes;
struct PAGE *subTrRoot;
long subTrPtr;
{
    register int i, k, leftSide, rightSide;
    struct PAGE temp;
    register long son;
    int tCmp;

    if(subTrPtr == (int) NULL)
        adjFlag = FALSE;
    else {
        leftSide = (int) NULL;
        rightSide = subTrRoot->nodSiz-1;
        do {
            k = (leftSide + rightSide) / 2;
            tCmp = ndxCmp(keyDes, keyValue, subTrRoot->nodElmt[k].srchVal);
            if(tCmp <= (int) NULL)
                rightSide = k - 1;
            if(tCmp >= (int) NULL)
                leftSide = k + 1;
        } while (rightSide >= leftSide);

        son = (rightSide < (int) NULL) ? subTrRoot->lftSon:
                                       subTrRoot->nodElmt[rightSide].child;
        if(tCmp == (int) NULL){
            btResult = subTrRoot->nodElmt[k].spare;
            Index[keyDes].nKeys--;
            if(son == (int) NULL){
                subTrRoot->nodSiz--;
                adjFlag = subTrRoot->nodSiz < MINKEYS;
                for(i = k; i < subTrRoot->nodSiz; i++)
                    itemcp(keyDes,&subTrRoot->nodElmt[i],
                                                &subTrRoot->nodElmt[i+1]);
            } else {
                if(getRep(keyDes, son, subTrRoot, k))
                    return ERROR;
                if(adjFlag)
                    if(underflow(keyDes,subTrRoot,son,rightSide))
                        return ERROR;
            }
            if(lioWrtKey(keyDes, subTrPtr, subTrRoot))
                return ERROR;
        } else {
            if(son) if(lioRdKey(keyDes, son, &temp))
                return ERROR;
            if(btDelete(keyDes, &temp, son))
                return ERROR;

            if(adjFlag){
                if(underflow(keyDes, subTrRoot, son, rightSide))
                    return ERROR;
                if(lioWrtKey(keyDes, subTrPtr, subTrRoot))
                    return ERROR;
            }
        }
    }
    return (int) NULL;
}
/*
 *    underflow - merge two adjacent pages if one underflows.
 */
LOCAL int underflow(keyDes, ancestor, subTrRoot, srchPosn)
struct PAGE *ancestor;
long subTrRoot;
int srchPosn, keyDes;
{
    struct PAGE temp, brother;
    register int i, k, brthrSiz, ancstrSiz;
    register long b;

    if(lioRdKey(keyDes, subTrRoot, &temp))
        return ERROR;
    ancstrSiz = ancestor->nodSiz;
    if(srchPosn + 1 < ancstrSiz){
        srchPosn++;
        b = ancestor->nodElmt[srchPosn].child;
        if(lioRdKey(keyDes, b, &brother))
            return ERROR;
        brthrSiz = brother.nodSiz;
        k = (brthrSiz - MINKEYS + 1) / 2;
        itemcp(keyDes, &temp.nodElmt[MINKEYS-1],
                                        &ancestor->nodElmt[srchPosn]);
        temp.nodElmt[MINKEYS-1].child = brother.lftSon;

        if(k > (int) NULL){
            for(i = 0; i < k-1; i++)
                itemcp(keyDes, &temp.nodElmt[i + MINKEYS], &brother.nodElmt[i]);
            itemcp(keyDes, &ancestor->nodElmt[srchPosn],
                                                    &brother.nodElmt[k-1]);
            ancestor->nodElmt[srchPosn].child = b;
            brother.lftSon = brother.nodElmt[k-1].child;
            brthrSiz -= k;
            for(i = 0; i < brthrSiz; i++)
                itemcp(keyDes, &brother.nodElmt[i], &brother.nodElmt[i + k]);
            brother.nodSiz = brthrSiz;
            temp.nodSiz = MINKEYS-1 + k;
            adjFlag = FALSE;
            if(lioWrtKey(keyDes, b, &brother))
                return ERROR;
        } else {
            for(i = 0; i < MINKEYS; i++)
                itemcp(keyDes, &temp.nodElmt[i + MINKEYS], &brother.nodElmt[i]);
            for(i = srchPosn; i < ancstrSiz-1; i++)
                itemcp(keyDes, &ancestor->nodElmt[i],
                                                    &ancestor->nodElmt[i+1]);
            temp.nodSiz = MAXKEYS;
            ancestor->nodSiz = ancstrSiz-1;
            if(lioFree(keyDes, b))
                return ERROR;
            adjFlag = ancestor->nodSiz < MINKEYS;
        }
        if(lioWrtKey(keyDes, subTrRoot, &temp))
            return ERROR;
    } else {
        b = (srchPosn == (int) NULL) ? ancestor->lftSon :
                                     ancestor->nodElmt[srchPosn-1].child;
        if(lioRdKey(keyDes, b, &brother))
            return ERROR;
        brthrSiz = brother.nodSiz + 1;
        k = (brthrSiz - MINKEYS) / 2;
        if(k > (int) NULL){
            for(i = MINKEYS - 2; i >= (int) NULL; i--)
            itemcp(keyDes, &temp.nodElmt[i + k], &temp.nodElmt[i]);
                itemcp(keyDes, &temp.nodElmt[k-1],
                                                &ancestor->nodElmt[srchPosn]);
            temp.nodElmt[k-1].child = temp.lftSon;
            brthrSiz -= k;
            for(i = k - 2; i >= (int) NULL; i--)
                itemcp(keyDes, &temp.nodElmt[i], &brother.nodElmt[i+brthrSiz]);
            temp.lftSon = brother.nodElmt[brthrSiz-1].child;
            itemcp(keyDes,&ancestor->nodElmt[srchPosn],
                                                &brother.nodElmt[brthrSiz-1]);
            ancestor->nodElmt[srchPosn].child = subTrRoot;
            brother.nodSiz = brthrSiz - 1;
            temp.nodSiz = MINKEYS - 1 + k;

            adjFlag = FALSE;
            if(lioWrtKey(keyDes, subTrRoot, &temp))
                return ERROR;
        } else {
            itemcp(keyDes, &brother.nodElmt[brthrSiz-1],
                                                &ancestor->nodElmt[srchPosn]);
            brother.nodElmt[brthrSiz-1].child = temp.lftSon;
            for(i = 0; i < MINKEYS-1; i++)
                itemcp(keyDes, &brother.nodElmt[i+brthrSiz], &temp.nodElmt[i]);
            brother.nodSiz = MAXKEYS;
            ancestor->nodSiz = ancstrSiz-1;
            if(lioFree(keyDes, subTrRoot))
                return ERROR;
            adjFlag = ancestor->nodSiz < MINKEYS;
        }
        if(lioWrtKey(keyDes, b, &brother))
            return ERROR;
    }
    return (int) NULL;
}
/*
 *    getRep - look for the best candidate to replace the item deleted.
 */
LOCAL int getRep(keyDes, curNode, subTrRoot, k)
int         keyDes, k;
long        curNode;
struct PAGE *subTrRoot;
{
    struct PAGE temp;
    long son;

    if(lioRdKey(keyDes, curNode, &temp))
        return ERROR;
    son = temp.nodElmt[temp.nodSiz-1].child;
    if(son){
        if(getRep(keyDes, son, subTrRoot, k))
            return ERROR;
        if(adjFlag){
            if(underflow(keyDes, &temp, son,  temp.nodSiz-1))
                return ERROR;
            if(lioWrtKey(keyDes, curNode, &temp))
                return ERROR;
        }
    } else {
        temp.nodSiz--;
        temp.nodElmt[temp.nodSiz].child = subTrRoot->nodElmt[k].child;
        itemcp(keyDes, &subTrRoot->nodElmt[k], &temp.nodElmt[temp.nodSiz]);
        adjFlag = temp.nodSiz < MINKEYS;
        if(lioWrtKey(keyDes, curNode, &temp))
            return ERROR;
    }
    return (int) NULL;
}
/*
 *   simple entry point prolog
 */
LOCAL int prolog(keyDes)
int keyDes;
{
    if(!initialized)
        lioInit();
    if(lioSize(keyDes) == LNULL)
        return ERROR;
    return (int) NULL;
}
/***************** End of all of the low level b-tree drivers ***************/
/*
 *  get size of key
 */
ENTRY int lioSizeKey( int keyDes ) {
    return( kkSize( keyDes ) );
}
/*
 *  get type of key
 */
ENTRY int lioTypeKey( int keyDes ) {
    return( kkType( keyDes ) );
}
/*
 *  insert a new key into the tree
 */
LOCAL long trInsert(keyDes, insKey, insRcdNo)
int  keyDes;
char *insKey;
long insRcdNo;
{
    long makNod();
    register long q;

    if(!initialized)
        lioInit();
    hasOver = FALSE;
    btResult = NOTFOUND;
    memcpy(keyValue, insKey, Index[keyDes].keySize);
    if(Index[keyDes].hasDup)
        ltoc(LNULL, keyValue + Index[keyDes].keySize - 3);
    keyRcdNo = insRcdNo;
    memset((char *)Index[keyDes].stack, (char) NULL,
                                       sizeof(Index[keyDes].stack));
    memset(Index[keyDes].matchName, (char) NULL,
                                  sizeof(Index[keyDes].matchName));
    Index[keyDes].stack[0] = Index[keyDes].treeRoot;
    Index[keyDes].stackPtr = 1;

    if(Index[keyDes].numRecs)
        if(lioRdKey(keyDes,Index[keyDes].treeRoot,&rootPg))
            return LERROR;
    if(btSearch(keyDes,&rootPg,&currItem,Index[keyDes].treeRoot))
        return LERROR;

    if(adjFlag){
        if(Index[keyDes].numRecs)
            if(lioWrtKey(keyDes, Index[keyDes].treeRoot, &rootPg))
                return LERROR;
        q = Index[keyDes].treeRoot;
        if((Index[keyDes].treeRoot=makNod(keyDes,&rootPg))==LERROR)
            return LERROR;
        rootPg.nodSiz = 1;
        rootPg.lftSon = q;
        itemcp(keyDes, &rootPg.nodElmt[0], &currItem);
        if(lioWrtKey(keyDes, Index[keyDes].treeRoot, &rootPg))
            return LERROR;
    }
    nodFree();
    return btResult;
}
/*
 *  search  look for a key in the tree.
 *          85.05.21    fixed bug to do with search for a key after
 *                      first in a series of duplicates deleted.
 */
ENTRY long lioSearch(keyDes, insKey)
int  keyDes;
char *insKey;
{
    char tKey[MAXKEYSIZE];
    long recno, lioNext();

    if(prolog(keyDes))
        return LNULL;
    amInserting = FALSE;
    if(trInsert(keyDes, insKey, (long) NULL) == LERROR)
        return LERROR;
    if(btResult == LNULL && Index[keyDes].hasDup){
        if((recno = lioNext(keyDes, tKey)) == LERROR)
            return LERROR;
        if(!ndxCmp1(keyDes, insKey, tKey))
            return recno;
        else
            return LNULL;
    }
    return btResult;
}
/*
 *  prepare to insert a new key
 */
ENTRY long lioInsert(keyDes, insKey, insRcdNo)
int  keyDes;
char *insKey;
long insRcdNo;
{
    register long result;

    amInserting = TRUE;
    treeShuffle = FALSE;
    result = trInsert(keyDes, insKey, insRcdNo);
    amInserting = FALSE;
    /*
     *  this search is necessary to maintain the currency, but only
     *  if a page has split
     */
    if(treeShuffle)
        if(trInsert(keyDes, insKey, (long) NULL) == LERROR)
            return LERROR;

    return result;
}
/*
 *    delete - remove a key value from the index tree
 */
ENTRY long lioDelete(keyDes, delKey)
int  keyDes;
char *delKey;
{
    register long q;

    if(prolog(keyDes))
        return LNULL;
    btResult = NOTFOUND;

    if(delKey){
        memcpy(keyValue, delKey, MAXKEYSIZE);
        if( (q = lioSearch(keyDes, keyValue) ) <= (long) NULL)
            return q;
    }
    memcpy(keyValue, Index[keyDes].matchName, Index[keyDes].keySize);
    memset((char *)Index[keyDes].stack, (char) NULL,sizeof(Index[keyDes].stack));
    Index[keyDes].stackPtr = (int) NULL;
    if(lioRdKey(keyDes, Index[keyDes].treeRoot, &rootPg))
        return LERROR;
    if(btDelete(keyDes, &rootPg, Index[keyDes].treeRoot))
        return LERROR;
    if(adjFlag)
        if(rootPg.nodSiz == (int) NULL){
            q = Index[keyDes].treeRoot;
            Index[keyDes].treeRoot = rootPg.lftSon;
            if(lioFree(keyDes, q))
                return LERROR;
        } else if(lioWrtKey(keyDes, Index[keyDes].treeRoot, &rootPg))
            return LERROR;

    nodFree();
    return btResult;
}
/*
 *    find the previous key in the tree
 */
ENTRY long lioPrev(keyDes, foundKey)
register int keyDes;
char         *foundKey;
{
    struct PAGE buf;
    register int i;
    short stackVal;
    long lioFirst();

    if(prolog(keyDes))
        return LNULL;

    if(lioRdKey(keyDes, Index[keyDes].stack[Index[keyDes].stackPtr-1], &buf))
        return LERROR;
    for(i = 0; i < buf.nodSiz; i++)
        if(ndxCmp(keyDes, buf.nodElmt[i].srchVal, Index[keyDes].matchName)
                                                                >= (int) NULL)
            break;

    if(buf.nodElmt[i].child == LNULL){
        if(i != (int) NULL){
            memcpy(Index[keyDes].matchName,
                     buf.nodElmt[i-1].srchVal, Index[keyDes].keySize);
            memcpy(foundKey,
                     buf.nodElmt[i-1].srchVal, kkSize(keyDes) );
            nodFree();
            return buf.nodElmt[i-1].spare;
        } else {
            loop
                if(--Index[keyDes].stackPtr == (int) NULL)
                    break;
                if(lioRdKey(keyDes,
                            Index[keyDes].stack[Index[keyDes].stackPtr-1],&buf))
                    return LERROR;

                for(i = buf.nodSiz-1; i >= (int) NULL; i--)
                    if(ndxCmp(keyDes,buf.nodElmt[i].srchVal,
                                    Index[keyDes].matchName) < (int) NULL){
                        memcpy(Index[keyDes].matchName,
                                buf.nodElmt[i].srchVal, Index[keyDes].keySize);
                        memcpy(foundKey, buf.nodElmt[i].srchVal,
                                                                kkSize(keyDes) );
                        nodFree();
                        return buf.nodElmt[i].spare;
                    }
            endLoop
            if(lioFirst(keyDes, foundKey) < LNULL)
                return LERROR;
            nodFree();
            return btResult;
        }
    } else {
        stackVal = Index[keyDes].stackPtr;
        Index[keyDes].stack[stackVal] =
                    (i != (int) NULL) ? buf.nodElmt[i-1].child : buf.lftSon;
        Index[keyDes].stackPtr++;
        loop
            if(lioRdKey(keyDes,Index[keyDes].stack[Index[keyDes].stackPtr-1],
                                                    &buf))
                return LERROR;
            stackVal = Index[keyDes].stackPtr;
            if((Index[keyDes].stack[stackVal]=
                            buf.nodElmt[buf.nodSiz-1].child) == (int) NULL)
                break;
            Index[keyDes].stackPtr++;
        endLoop
        memcpy(Index[keyDes].matchName,
              buf.nodElmt[buf.nodSiz-1].srchVal, Index[keyDes].keySize);
        memcpy(foundKey,
                    buf.nodElmt[buf.nodSiz-1].srchVal, kkSize(keyDes) );
        nodFree();
        return buf.nodElmt[buf.nodSiz-1].spare;
    }
}
/*
 *  Search Partial  Do a search looking for the key that is greater
 *                  than or equal to the search key. If the end of
 *                  the tree is hit, return the last key in the tree.
 */
ENTRY long lioSrPrtl(keyDes, srchKey, foundKey)
int keyDes;
char *srchKey, *foundKey;
{
    struct PAGE  srchPg;
    register int i;
    short        stackVal;
    long         save;

    if(prolog(keyDes))
        return LNULL;

    (void) trInsert(keyDes, srchKey, (long) NULL);
    /*
     *  we have the key before the one greater than or equal to the search
     *  key. Look for the next one in the tree.
     */
    do {
        if(lioRdKey(keyDes, Index[keyDes].stack[Index[keyDes].stackPtr-1],
                                                        &srchPg))
            return LERROR;
        for(i = 0; i < srchPg.nodSiz; i++)
            if(ndxCmp1(keyDes, srchKey, srchPg.nodElmt[i].srchVal) <=
                                                            (int) NULL){
                memcpy(Index[keyDes].matchName,
                     srchPg.nodElmt[i].srchVal, Index[keyDes].keySize);
                memcpy(foundKey, srchPg.nodElmt[i].srchVal,
                                                     kkSize(keyDes) );
                nodFree();
                return srchPg.nodElmt[i].spare;
            }

        stackVal = Index[keyDes].stackPtr--;
        Index[keyDes].stack[stackVal] = (int) NULL;
    } while(Index[keyDes].stackPtr > (int) NULL);
    /*
     *   looking for key > largest key in tree, return largest key.
     */
    if( (save = lioLast(keyDes, foundKey)) < LNULL)
        return LERROR;
    memcpy(Index[keyDes].matchName, foundKey, kkSize(keyDes) );

    nodFree();
    return save;
}
/*
 *    get next key - return the key just greater than the one previously
 *           returned by a search or searchPartial call.
 */
ENTRY long lioNext(keyDes, foundKey)
int  keyDes;
char *foundKey;
{
    struct PAGE  buf;
    register int i;
    short        stackVal;

    if(prolog(keyDes))
        return LNULL;
    if(lioRdKey(keyDes, Index[keyDes].stack[Index[keyDes].stackPtr-1], &buf))
        return LERROR;
    for(i = buf.nodSiz-1; i >= (int) NULL; i--)
    if(ndxCmp(keyDes, buf.nodElmt[i].srchVal, Index[keyDes].matchName)
                                                                <= (int) NULL)
        break;

    if(buf.nodElmt[i].child == LNULL){
        if(i < buf.nodSiz-1){
            memcpy(Index[keyDes].matchName,
                     buf.nodElmt[i+1].srchVal, Index[keyDes].keySize);
            memcpy(foundKey,
                         buf.nodElmt[i+1].srchVal, kkSize(keyDes) );
            nodFree();
            return buf.nodElmt[i + 1].spare;
        } else {
            loop
                if(--Index[keyDes].stackPtr == (int) NULL)
                    break;
                if(lioRdKey(keyDes,
                            Index[keyDes].stack[Index[keyDes].stackPtr-1],&buf))
                    return LERROR;
                for(i = 0; i < buf.nodSiz; i++)
                    if(ndxCmp(keyDes,buf.nodElmt[i].srchVal,
                                    Index[keyDes].matchName) > (int) NULL){
                        memcpy(Index[keyDes].matchName,
                          buf.nodElmt[i].srchVal, Index[keyDes].keySize);
                        memcpy(foundKey, buf.nodElmt[i].srchVal,
                                                     kkSize(keyDes) );
                        nodFree();
                        return buf.nodElmt[i].spare;
                    }
            endLoop
            if(lioLast(keyDes, foundKey) < LNULL)
                return LERROR;
            nodFree();
            return btResult;
        }
    } else {
        stackVal = Index[keyDes].stackPtr;
        Index[keyDes].stack[stackVal] = buf.nodElmt[i].child;
        Index[keyDes].stackPtr++;
        loop
            if(lioRdKey(keyDes,Index[keyDes].stack[Index[keyDes].stackPtr-1],&buf))
                return LERROR;
            stackVal = Index[keyDes].stackPtr;
            if((Index[keyDes].stack[stackVal] = buf.lftSon) == (int) NULL)
                break;
            Index[keyDes].stackPtr++;
        endLoop
        memcpy(Index[keyDes].matchName, buf.nodElmt[0].srchVal,
               Index[keyDes].keySize);
        memcpy(foundKey, buf.nodElmt[0].srchVal, kkSize(keyDes) );
        nodFree();
        return buf.nodElmt[0].spare;
    }
}
/*
 *   first - find the first key in a tree
 */
ENTRY long lioFirst(keyDes, fk)
int  keyDes;
char *fk;
{
    if(prolog(keyDes))
        return LNULL;
    memset((char *) Index[keyDes].stack, (char) NULL,sizeof(Index[keyDes].stack));
    memset(Index[keyDes].matchName, (char) NULL, Index[keyDes].keySize);
    Index[keyDes].stackPtr = (int) NULL;
    btResult = NOTFOUND;
    if(btFirst(keyDes, Index[keyDes].treeRoot, fk))
        return LERROR;
    nodFree();
    return btResult;
}
/*
 *   descend the leftmost branches of a tree looking for the smallest leaf.
 */
LOCAL int btFirst(keyDes, rootNod, fk)
int  keyDes;
long rootNod;
char *fk;
{
    register long son;
    short         stackVal;
    struct PAGE   nextPage;

    if(rootNod != LNULL){
        if(lioRdKey(keyDes, rootNod, &nextPage))
            return ERROR;
        son = nextPage.lftSon;
        stackVal = Index[keyDes].stackPtr;
        Index[keyDes].stack[stackVal] = rootNod;
        Index[keyDes].stackPtr++;
        if(son){
            if(btFirst(keyDes, son, fk))
                return ERROR;
        } else {
            memcpy(Index[keyDes].matchName, nextPage.nodElmt[0].srchVal,
                                                 Index[keyDes].keySize);
            memcpy(fk, nextPage.nodElmt[0].srchVal, kkSize(keyDes) ) ;
            btResult = nextPage.nodElmt[0].spare;
        }
    }
    return (int) NULL;
}
/*
 *   last - find the last key in a tree
 */
ENTRY long lioLast(keyDes, fk)
int  keyDes;
char *fk;
{
    if(prolog(keyDes))
        return LNULL;
    memset((char *) Index[keyDes].stack, (char) NULL,sizeof(Index[keyDes].stack));
    memset(Index[keyDes].matchName, (char) NULL, Index[keyDes].keySize);
    Index[keyDes].stackPtr = (int) NULL;
    btResult = NOTFOUND;
    if(btLast(keyDes, Index[keyDes].treeRoot, fk))
        return LERROR;
    nodFree();
    return btResult;
}
/*
 *   descend the rightmost branches of a tree looking for the largest leaf.
 */
LOCAL int btLast(keyDes, rootNod, fk)
int keyDes;
long rootNod;
char *fk;
{
    register long son;
    short stackVal;
    struct PAGE nextPage;

    if(rootNod != LNULL){
        if(lioRdKey(keyDes, rootNod, &nextPage))
            return ERROR;
        son = nextPage.nodElmt[nextPage.nodSiz-1].child;
        stackVal = Index[keyDes].stackPtr;
        Index[keyDes].stack[stackVal] = rootNod;
        Index[keyDes].stackPtr++;
        if(son){
            if(btLast(keyDes, son, fk))
                return ERROR;
        } else {
            memcpy(Index[keyDes].matchName,
                 nextPage.nodElmt[nextPage.nodSiz-1].srchVal,
                     Index[keyDes].keySize);
            memcpy(fk, nextPage.nodElmt[nextPage.nodSiz-1].srchVal,
                     kkSize(keyDes) );
            btResult = nextPage.nodElmt[nextPage.nodSiz-1].spare;
        }
    }
    return (int) NULL;
}
/*
 *  find the key with a specific data record number. Return the record
 *  number if found, LNULL otherwise.
 */
ENTRY long lioSerNum(keyDes, key, num)
int keyDes;
char *key;
long num;
{
    char buf[MAXKEYSIZE+1];
    long found, prev, lioSearch(), lioNext();

    if( (prev = found = lioSearch(keyDes, key)) == LERROR)
        return LNULL;

    if( !Index[keyDes].hasDup)
        return (found == num) ? found : LNULL;

    loop
        if(found == num)
            return num;
        if( (found = lioNext(keyDes, buf)) == LERROR)
            return LNULL;
        if( ndxCmp1( keyDes, buf, key) != (int) NULL)
            break;
        if( found == prev )
            break;
        prev = found;
    endLoop

    return LNULL;
}
/*
 *  comment this include out if treeprint is not needed'
 *  Only people doing fairly heavy mods to the actual btree routines
 *  should need this.
 */
/* #include "lioTrPr.c" */
/* ------------------------- Modifiable functions ------------------------- */

/*
 *   Index compare
 *   - this function compares two values (Left & Right) and returns:
 *  1 if L > R
 *  0 if L == R and
 *     -1 if L < R
 *
 *   If you want to use 'custom' keys, such as integer or float
 *   change this function.
 */
LOCAL int ndxCmp(keyDes, left, right)
int  keyDes;
char *left,
     *right;
{
   register int i;
   register long l, r;

   i = ndxCmp1(keyDes, left, right);
   if(!Index[keyDes].hasDup || i) return i;

   i = Index[keyDes].keySize - 3;
   if(amInserting && (ctol(left + i) == (long) NULL)){
      ltoc(++Index[keyDes].dupNum, left + Index[keyDes].keySize - 3);
      return 1;
   }
   l = ctol(left + i);
   r = ctol(right + i);
   if(l < r) return -1;
   if(l > r) return 1;

   return (int) NULL;
}
/*
 *  compare the two strings given for a maximum of <len> bytes
 */

ENTRY int ndxCmp1(keyDes, left, right)
int keyDes;
char *left, *right;
{
    int         i, len;
    int         i_left, i_right;
    long        l_left, l_right;
    short       s_left, s_right;
    double      d_left, d_right, dt;
   
    switch( kkType( keyDes ) ) {
        case tSTRING :
            len = kkSize( keyDes );
            for(i = 0; i < len; i++){
        		if( (unsigned char)left[i] < (unsigned char)right[i] )
        		    return( -1 );
        		if( (unsigned char)left[i] > (unsigned char)right[i] )
        		    return( 1 );
        	}
        	return( 0 );

        case tDECIMAL :
            memcpy( &d_left, left, zDECIMAL );
            memcpy( &d_right, right, zDECIMAL );
            dt = d_left - d_right;
            if( dt < 0.0 ) return( -1 );
            if( dt > 0.0 ) return( 1 );
            return( 0 );
            
        case tSHORT :
            memcpy( &s_left, left, zSHORT );
            memcpy( &s_right, right, zSHORT );
            return( (int)(s_left - s_right) );
            
        case tUSHORT :
            memcpy( &s_left, left, zUSHORT );
            memcpy( &s_right, right, zUSHORT );
            return( (int)((unsigned short)s_left - (unsigned short)s_right) );
            
        case tINTEGER :
            memcpy( &i_left, left, zINTEGER );
            memcpy( &i_right, right, zINTEGER );
            return( (i_left - i_right) );
            
        case tUINTEGER :
            memcpy( &i_left, left, zUINTEGER );
            memcpy( &i_right, right, zUINTEGER );
            return( ((unsigned int)i_left - (unsigned int)i_right) );
            
        case tLONG :
            memcpy( &l_left, left, zLONG );
            memcpy( &l_right, right, zLONG );
            return( (l_left - l_right) );
        
        case tULONG :
            memcpy( &l_left, left, zULONG );
            memcpy( &l_right, right, zULONG );
            return( ((unsigned long)l_left - (unsigned long)l_right) );
        
        case tTRIPLE :
            l_left = ctol( left );
            l_right = ctol( right );
            return( (l_left - l_right) );

        default :
            printf( "ndxCmp1(): Unknown Key Type `%d'!\n", kkType(keyDes) );
            exit( -1 );
    }
}
